/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#define L1_CACHE_LINE_MAX 128 /* largest I've seen (Pentium 4) */
static size_t l1_cache_line_size = L1_CACHE_LINE_MAX;
static pthread_key_t mog_rbuf_key; /* for small socket/pipe reads (headers) */
static pthread_key_t mog_fsbuf_key; /* large for filesystem I/O */
static size_t FSBUF_SIZE = BUFSIZ; /* trust stdio, BUFSIZ=8192 on glibc */

#define MOG_MASK(align)        (~((size_t)align - 1))
#define MOG_ALIGN(align,val)   (((val) + (size_t)align - 1) & MOG_MASK(align))

/*
 * Just in case the system page size is larger than BUFSIZ,
 * crank up FSBUF_SIZE to match since mmap() will internally
 * round up the allocation anyways
 */
#ifdef _SC_PAGE_SIZE
static void fsbuf_size_page_align(void)
{
	long page_size = sysconf(_SC_PAGE_SIZE);

	if (page_size > 0) {
		if (page_size > FSBUF_SIZE)
			FSBUF_SIZE = page_size;
		else
			FSBUF_SIZE = MOG_ALIGN(page_size, FSBUF_SIZE);
	}
}
#else
# define fsbuf_size_page_align() for(;0;)
#endif /* _SC_PAGE_SIZE */


#ifdef _SC_LEVEL1_DCACHE_LINESIZE
static size_t l1_cache_line_size_detect(void)
{
	long tmp = sysconf(_SC_LEVEL1_DCACHE_LINESIZE);

	if (tmp > 0 && tmp <= L1_CACHE_LINE_MAX) return (size_t)tmp;

	return L1_CACHE_LINE_MAX;
}
#else
# define l1_cache_line_size_detect() (L1_CACHE_LINE_MAX)
#endif /* _SC_LEVEL1_DCACHE_LINESIZE */


static void fsbuf_free(void *);

__attribute__((constructor)) static /* FIXME for non-gcc */
void mog_alloc_init(void)
{
	fsbuf_size_page_align();
	l1_cache_line_size = l1_cache_line_size_detect();
	CHECK(int, 0, pthread_key_create(&mog_rbuf_key, free));
	CHECK(int, 0, pthread_key_create(&mog_fsbuf_key, fsbuf_free));
}


void mog_free_and_null(void *ptrptr)
{
	void **tmp = ptrptr;

	free(*tmp);
	*tmp = NULL;
}

_Noreturn void mog_oom(void)
{
	write(STDERR_FILENO, "OOM\n", 4);
	syslog(LOG_CRIT, "Out of memory, aborting");
	abort();
}


/*
 * Cache alignment is important for sub-pagesized allocations
 * that can be bounced between threads.  We round up the
 * allocation to the cache size
 */
void *mog_cachealign(size_t size)
{
	void *ptr;
	int err = posix_memalign(&ptr, l1_cache_line_size, size);

	switch (err) {
	case 0: return ptr;
	case ENOMEM: mog_oom();
	}
	die("posix_memalign failed: %s\n", strerror(err));
}


/* allocates a new mog_rbuf of +size+ bytes */
struct mog_rbuf *mog_rbuf_new(size_t size)
{
	struct mog_rbuf *rbuf = mog_cachealign(size + sizeof(struct mog_rbuf));

	assert(size > 0 && "tried to allocate a zero-byte mog_rbuf");
	rbuf->rsize = size;

	return rbuf;
}

MOG_NOINLINE static struct mog_rbuf *
rbuf_replace(struct mog_rbuf *rbuf, size_t size)
{
	free(rbuf); /* free(NULL) works on modern systems */
	rbuf = mog_rbuf_new(size);
	CHECK(int, 0, pthread_setspecific(mog_rbuf_key, rbuf));

	return rbuf;
}

/*
 * retrieves the per-thread rbuf belonging to the current thread,
 * ensuring it is at least capable of storing the specified size
 */
struct mog_rbuf *mog_rbuf_get(size_t size)
{
	struct mog_rbuf *rbuf = pthread_getspecific(mog_rbuf_key);

	if (rbuf && rbuf->rsize >= size) return rbuf;

	return rbuf_replace(rbuf, size);
}

/* ensures a given rbuf is no longer associated with the current thread */
struct mog_rbuf *mog_rbuf_defer(struct mog_rbuf *rbuf)
{
	struct mog_rbuf *cur = pthread_getspecific(mog_rbuf_key);

	if (cur == rbuf)
		CHECK(int, 0, pthread_setspecific(mog_rbuf_key, NULL));

	return rbuf;
}

void mog_rbuf_free(struct mog_rbuf *rbuf)
{
	assert(((rbuf == NULL) ||
	       (pthread_getspecific(mog_rbuf_key) != rbuf)) &&
	       "trying to free undeferred rbuf");
	free(rbuf);
}

void mog_rbuf_free_and_null(struct mog_rbuf **ptrptr)
{
	mog_rbuf_free(*ptrptr);
	*ptrptr = NULL;
}


#ifndef MAP_ANONYMOUS
#  define MAP_ANONYMOUS MAP_ANON
#endif

MOG_NOINLINE static void * fsbuf_new(void)
{
	int prot = PROT_READ | PROT_WRITE;
	int flags = MAP_ANONYMOUS | MAP_PRIVATE;
	int fd = -1;
	off_t offset = 0;
	void *ptr = mmap(NULL, FSBUF_SIZE, prot, flags, fd, offset);

	if (ptr == MAP_FAILED) mog_oom();

	CHECK(int, 0, pthread_setspecific(mog_fsbuf_key, ptr));

	return ptr;
}

/* retrieves the per-thread fsbuf and sets size to the value of FSBUF_SIZE */
void *mog_fsbuf_get(size_t *size)
{
	void *ptr = pthread_getspecific(mog_fsbuf_key);

	*size = FSBUF_SIZE;
	if (ptr) return ptr;

	return fsbuf_new();
}

/* destructor called at thread exit */
static void fsbuf_free(void *ptr)
{
	if (ptr) CHECK(int, 0, munmap(ptr, FSBUF_SIZE));
}
