/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "cfg.h"
#define THIS "cmogstored"
#define DESC "alternative mogstored implementation for MogileFS"
static char summary[] = THIS " -- "DESC;
const char *argp_program_bug_address = PACKAGE_BUGREPORT;
const char *argp_program_version = THIS" "PACKAGE_VERSION;
static sig_atomic_t sigchld_nr;
static sig_atomic_t do_exit;

#define CFG_KEY(f) -((int)offsetof(struct mog_cfg,f) + 1)
static struct argp_option options[] = {
	{ .name = "daemonize", .key = 'd',
	  .doc = "Daemonize" },
	{ .name = "config", .key = CFG_KEY(configfile),
	  .arg = "<file>",
          .doc = "Set config file (no default, unlike mogstored)" },
	{ .name = "httplisten", .key = CFG_KEY(httplisten),
	  .arg = "<ip:port>",
	  .doc = "IP/Port HTTP server listens on" },
	{ .name = "mgmtlisten", .key = CFG_KEY(mgmtlisten),
	  .arg = "<ip:port>",
	  .doc = "IP/Port management/sidechannel listens on" },
	{ .name = "docroot", .key = CFG_KEY(docroot),
	  .arg = "<path>",
	  .doc = "Docroot above device mount points.  "
	         "Defaults to "MOG_DEFAULT_DOCROOT
	},
	{ .name = "maxconns", .key = CFG_KEY(maxconns),
	  .arg = "<number>",
	  .doc = "Number of simultaneous clients to serve. "
	         "Default " MOG_STR(MOG_DEFAULT_MAXCONNS) },
	{ .name = "pidfile", .key = CFG_KEY(pidfile),
	  .arg = "<path>",
	  .doc = "path to PID file" },
	{ .name = "server", .key = CFG_KEY(server), .flags = OPTION_HIDDEN },
	{
	  /* hidden for now, don't break compat with Perl mogstored */
	  .name = "multi", .key = 'M', .flags = OPTION_HIDDEN
	},
	{ NULL }
};

static void new_cfg_or_die(const char *config)
{
	struct mog_cfg *cfg = mog_cfg_new(config);

	if (!cfg) die("invalid (or duplicate) config=%s\n", config);
	if (mog_cfg_load(cfg) == 0) return;

	die("failed to load config=%s (%s)\n",
	    config, errno ? strerror(errno) : "parser error");
}

static void maxconns_or_die(struct mog_cfg *cfg, const char *s)
{
	char *end;

	errno = 0;
	cfg->maxconns = strtoul(s, &end, 10);
	if (errno)
		die("failed to parse maxconns=%s (%s)\n", s, strerror(errno));
	if (*end)
		die("failed to parse maxconns=%s (invalid character: %c)\n",
		    s, *end);
}

static void addr_or_die(struct mog_addrinfo **dst, const char *key, char *s)
{
	*dst = mog_listen_parse(s);
	if (!*dst)
		die("failed to parse %s=%s\n", key, s);
}

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	struct mog_cfg *cfg = state->input;
	int rv = 0;

	switch (key) {
	case 'd': cfg->daemonize = 1; break;
	case CFG_KEY(docroot): cfg->docroot = xstrdup(arg); break;
	case CFG_KEY(pidfile): cfg->pidfile = xstrdup(arg); break;
	case CFG_KEY(configfile): new_cfg_or_die(arg); break;
	case CFG_KEY(maxconns): maxconns_or_die(cfg, arg); break;
	case CFG_KEY(httplisten):
		addr_or_die(&cfg->httplisten, "httplisten", arg);
		break;
	case CFG_KEY(mgmtlisten):
		addr_or_die(&cfg->mgmtlisten, "mgmtlisten", arg);
		break;
	case CFG_KEY(server):
		cfg->server = xstrdup(arg);
		mog_cfg_check_server(cfg);
		break;
	case 'M': mog_cfg_multi = true; break;
	case ARGP_KEY_ARG:
		argp_usage(state);
	case ARGP_KEY_END:
		break;
	default:
		rv = ARGP_ERR_UNKNOWN;
	}

	return rv;
}

static int open_pidfile(void)
{
	pid_t cur_pid = -1;
	int pid_fd = mog_pidfile_open(mog_cli.pidfile, &cur_pid);

	if (pid_fd >= 0)
		return pid_fd;
	if (errno == EAGAIN)
		die("already running on PID: %d\n", (int)cur_pid);
	else
		die("mog_pidfile_open failed: %s\n", strerror(errno));
	return -1;
}

MOG_NOINLINE static void setup(int argc, char *argv[])
{
	int pid_fd = -1;
	static struct argp argp = { options, parse_opt, NULL, summary };
	int mask = 0;

	argp_parse(&argp, argc, argv, 0, NULL, &mog_cli);
	mog_cfg_validate_or_die(&mog_cli);
	mog_cfg_svc_start_or_die(&mog_cli);

	if (mog_cli.pidfile) pid_fd = open_pidfile();

	/* TODO: make logging configurable */
	{
		openlog(THIS, LOG_ODELAY|LOG_PID, LOG_DAEMON);
		mask |= LOG_MASK(LOG_EMERG);
		mask |= LOG_MASK(LOG_ALERT);
		mask |= LOG_MASK(LOG_CRIT);
		mask |= LOG_MASK(LOG_ERR);
		mask |= LOG_MASK(LOG_WARNING);
		mask |= LOG_MASK(LOG_NOTICE);
		mask |= LOG_MASK(LOG_INFO);
		/* mask |= LOG_MASK(LOG_DEBUG); */
		setlogmask(mask);
	}

	if (mog_cli.daemonize) daemon(0, 0);

	if (pid_fd >= 0 && mog_pidfile_commit(pid_fd) < 0)
		syslog(LOG_ERR,
		       "failed to write pidfile(%s): %m. continuing...",
		       mog_cli.pidfile);
}

/* Hash iterator function */
static bool svc_start_each(void *svcptr, void *qptr)
{
	struct mog_svc *svc = svcptr;
	struct mog_queue *q = qptr;
	struct mog_accept *ac;

	mog_svc_scandev(svc, mog_dev_mkusage);
	svc->queue = q;

	if (svc->mgmt_fd >= 0) {
		ac = mog_accept_init(svc->mgmt_fd, svc, mog_mgmt_post_accept);
		mog_thrpool_start(&ac->thrpool, 8, mog_accept_loop, ac);
	}

	/* TODO: support HTTP */
	if (svc->http_fd >= 0) {
		mog_close(svc->http_fd);
		svc->http_fd = -1;
	}
	if (svc->httpget_fd >= 0) {
		mog_close(svc->httpget_fd);
		svc->httpget_fd = -1;
	}

	return true;
}

static bool svc_quit_each(void *svcptr, void *ignored)
{
	struct mog_svc *svc = svcptr;
	struct mog_fd *mfd;
	struct mog_accept *ac;

	if (svc->mgmt_fd >= 0) {
		mfd = mog_fd_get(svc->mgmt_fd);
		ac = &mfd->as.accept;
		mog_thrpool_quit(&ac->thrpool);
	}

	return true;
}

_Noreturn static void cmogstored_exit(void)
{
	/* call atexit() handlers and make valgrind happy */
	mog_svc_each(svc_quit_each, NULL);
	mog_fdmap_stop_queues();
	exit(EXIT_SUCCESS);
}

static void cmogstored_wakeup_handler(int signum)
{
	switch (signum) {
	case SIGCHLD: ++sigchld_nr; break;
	case SIGQUIT:
	case SIGTERM:
	case SIGINT:
		do_exit = signum;
	}
	mog_notify(MOG_NOTIFY_SIGNAL);
}

static void cmogstored_siginit(void)
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(struct sigaction));
	CHECK(int, 0, sigemptyset(&sa.sa_mask) );

	sa.sa_handler = SIG_IGN;
	CHECK(int, 0, sigaction(SIGPIPE, &sa, NULL));

	sa.sa_handler = cmogstored_wakeup_handler;
	CHECK(int, 0, sigaction(SIGTERM, &sa, NULL));
	CHECK(int, 0, sigaction(SIGINT, &sa, NULL));

	/* TODO: graceful shutdown, upgrade, reload */
	CHECK(int, 0, sigaction(SIGQUIT, &sa, NULL));
	CHECK(int, 0, sigaction(SIGHUP, &sa, NULL));
	CHECK(int, 0, sigaction(SIGUSR1, &sa, NULL));
	CHECK(int, 0, sigaction(SIGUSR2, &sa, NULL));

	sa.sa_flags = SA_NOCLDSTOP;
	CHECK(int, 0, sigaction(SIGCHLD, &sa, NULL));
}

static void main_loop(struct mog_queue *q)
{
	for (;;) {
		mog_notify_wait();
		if (do_exit)
			cmogstored_exit();
		if (sigchld_nr > 0) {
			mog_mnt_refresh();
			mog_iostat_spawn(q);
			sigchld_nr = 0;
		}
	}
}

int main(int argc, char *argv[])
{
	struct mog_queue *q;

	mog_notify_init();
	mog_mnt_refresh();
	setup(argc, argv);

	q = mog_queue_new();
	mog_intr_disable();
	cmogstored_siginit();
	mog_thrpool_start(&q->thrpool, 8, mog_queue_loop, q);
	mog_svc_each(svc_start_each, q);
	mog_iostat_spawn(q);
	main_loop(q);

	return 0;
}
