/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
/* all path operations we support are relative to mog_svc.docroot */
#include "cmogstored.h"
#define MY_PATHMAX 80
#ifdef O_CLOEXEC
#  define MY_CLOEXEC O_CLOEXEC
#else
#  define MY_CLOEXEC 0
#endif

static int noatime_flags = O_RDONLY | O_NOATIME | MY_CLOEXEC;
static const int excl_flags = O_RDWR | O_EXCL | O_CREAT;

static inline int force_cloexec(int fd)
{
	if (MY_CLOEXEC == 0 && fd >= 0)
		set_cloexec_flag(fd, true);
	return fd;
}

/* we only use real *at syscalls, Gnulib workalikes aren't thread-safe */

#define GET_FSPATH(DST,SRC) do { \
	int rc = snprintf((DST), sizeof(DST), "%s%s", svc->docroot, (SRC)); \
	if (rc <= 0 || rc >= sizeof(DST)) { \
		errno = ENAMETOOLONG; \
		return -1; \
	} \
} while (0)

static int mog_statpath(struct mog_svc *svc, const char *path, struct stat *sb)
{
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	return stat(fspath, sb);
}

static int mog_openpath_read(struct mog_svc *svc, const char *path)
{
	int fd;
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	fd = open(fspath, noatime_flags, 0666);

	if (fd < 0 && errno != ENOENT && (noatime_flags & O_NOATIME)) {
		noatime_flags = O_RDONLY | MY_CLOEXEC;
		fd = open(fspath, noatime_flags, 0666);
	}
	return force_cloexec(fd);
}

static int mog_openpath_excl(struct mog_svc *svc, const char *path)
{
	int fd;
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	fd = open(fspath, excl_flags, 0666);

	return force_cloexec(fd);
}

#ifdef HAVE_FSTATAT
static int mog_statat(struct mog_svc *svc, const char *path, struct stat *sb)
{
	int rc = fstatat(svc->dirfd, path + 1, sb, 0);

	if (rc < 0 && errno == ENOSYS)
		return mog_statpath(svc, path, sb);
	return rc;
}
#else
#  define mog_statat mog_statpath
#endif

#ifdef HAVE_OPENAT
static int mog_openat_read(struct mog_svc *svc, const char *path)
{
	int rc = openat(svc->dirfd, path + 1, noatime_flags, 0666);

	if (rc < 0 && errno == ENOSYS)
		return mog_openpath_read(svc, path);

	if (rc < 0 && errno != ENOENT && (noatime_flags & O_NOATIME)) {
		noatime_flags = O_RDONLY | MY_CLOEXEC;
		rc = openat(svc->dirfd, path + 1, noatime_flags, 0666);
	}

	return force_cloexec(rc);
}

static int mog_openat_excl(struct mog_svc *svc, const char *path)
{
	int rc = openat(svc->dirfd, path + 1, excl_flags, 0666);

	if (rc < 0 && errno == ENOSYS)
		return mog_openpath_excl(svc, path);

	return force_cloexec(rc);
}

#else
#  define mog_openat_read mog_openpath_read
#  define mog_openat_excl mog_openpath_excl
#endif

static int mog_unlinkpath(struct mog_svc *svc, const char *path)
{
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath,path);
	return unlink(fspath);
}

#ifdef HAVE_UNLINKAT
static int mog_unlinkat(struct mog_svc *svc, const char *path)
{
	int rc = unlinkat(svc->dirfd, path + 1, 0);

	if (rc < 0 && errno == ENOSYS)
		return mog_unlinkpath(svc, path);
	return rc;
}
#else
#  define mog_unlinkat mog_unlinkpath
#endif

static int mog_renamepath(struct mog_svc *svc, const char *old, const char *new)
{
	char fsnew[MY_PATHMAX];
	char fsold[MY_PATHMAX];

	GET_FSPATH(fsold, old);
	GET_FSPATH(fsnew, new);
	return rename(fsold, fsnew);
}
#ifdef HAVE_RENAMEAT
static int mog_renameat(struct mog_svc *svc, const char *old, const char *new)
{
	int rc = renameat(svc->dirfd, old + 1, svc->dirfd, new + 1);

	if (rc < 0 && errno == ENOSYS)
		return mog_renamepath(svc, old, new);
	return rc;
}
#else
#  define mog_renameat mog_renamepath
#endif

int (*mog_stat)(struct mog_svc *, const char *, struct stat *) = mog_statat;
int (*mog_open_read)(struct mog_svc *, const char *) = mog_openat_read;
int (*mog_open_excl)(struct mog_svc *, const char *) = mog_openat_excl;
int (*mog_unlink)(struct mog_svc *, const char *) = mog_unlinkat;
int (*mog_rename)(struct mog_svc *, const char *, const char *) = mog_renameat;
