/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */

/*
 * iostat(1) -> global mountlist -> each mog_dev in each mog_svc
 */
#include "cmogstored.h"
#define MOG_IOSTAT_RBUF_SIZE (1024 - MOG_RBUF_OVERHEAD)

/* called after a stats line for a single device is complete */
void mog_iostat_line_done(struct mog_iostat *iostat, char *buf, size_t len)
{
	char *dev = buf + iostat->dev_tip;
	char util[8];

	assert(iostat->util_tip > iostat->dev_tip && "BUG: util_tip > dev_tip");
	assert((iostat->dev_len + iostat->dev_tip) < len && "dev_len overflow");
	assert((iostat->util_len + iostat->util_tip) <= len &&
	       "util_len overflow");

	dev[iostat->dev_len] = 0;
	if (iostat->util_len >= 7) {
		util[0] = '-';
		util[1] = 0;
	} else {
		memcpy(util, buf + iostat->util_tip, iostat->util_len);
		util[iostat->util_len] = 0;
	}

	mog_mnt_update_util(dev, util, sizeof(util));
}

/* called every second, after stats for each device line are out */
void mog_iostat_commit(void)
{
	mog_svc_each(mog_svc_devstats_broadcast, NULL);
}

static void iostat_advance(struct mog_iostat *iostat, char *buf, size_t len)
{
	size_t end = iostat->line_end + 1;

	if (iostat->line_end == 0) {
		syslog(LOG_ERR, "BUG: iostat line too long");
		assert(0 && "BUG: iostat line too long");
		abort();
	}

	assert((buf[iostat->line_end] == '\n' ||
	        buf[iostat->line_end] == '\0') &&
	       "iostat line_end is not LF or NUL");
	assert(iostat->offset > iostat->line_end &&
	       "iostat offset past line_end");
	assert(iostat->line_end < len && "iostat line_end farther than buf");

	iostat->offset -= end;
	memmove(buf, buf + end, len - end);

	if (iostat->dev_tip) iostat->dev_tip -= end;
	if (iostat->util_tip) iostat->util_tip -= end;

	iostat->line_end = 0;
}

static void iostat_close(struct mog_fd *mfd)
{
	assert(mfd->fd_type == MOG_FD_TYPE_IOSTAT && "bad fd_type");
	mog_close(mfd->fd);
}

void mog_iostat_queue_step(struct mog_fd *mfd)
{
	static const size_t capa = MOG_IOSTAT_RBUF_SIZE;
	struct mog_rbuf *rbuf = mog_rbuf_get(capa);
	ssize_t r;
	off_t off;
	struct mog_iostat *iostat = &mfd->as.iostat;
	size_t rbuf_used;

	assert(mfd->fd >= 0 && mfd->fd_type == MOG_FD_TYPE_IOSTAT &&
	       "bad iostat mfd");
retry:
	rbuf_used = 0;
	off = iostat->offset;
	r = read(mfd->fd, rbuf->rptr + off, capa - off);
	if (r > 0) {
		rbuf_used = r + off;
		switch (mog_iostat_parse(iostat, rbuf->rptr, rbuf_used)) {
		case MOG_IOSTAT_PARSER_ERROR:
			syslog(LOG_ERR, "iostat parser error");
			iostat_close(mfd);
			return;
		case MOG_IOSTAT_PARSER_DONE:
			assert(rbuf_used == iostat->offset &&
			       "iostat parser didn't finish");
			mog_iostat_init(iostat);
			goto retry;
		case MOG_IOSTAT_PARSER_WANTREAD:
			if (capa == iostat->offset)
				iostat_advance(iostat, rbuf->rptr, capa);
			goto retry;
		}
	} else if (r == 0) {
		/* iostat exits every 30s by default */
		iostat_close(mfd);
	} else {
		switch (errno) {
		case_EAGAIN:
			mog_idleq_push(iostat->queue, mfd, MOG_QEV_RD);
			return;
		case EINTR: goto retry;
		}
		syslog(LOG_ERR, "iostat read() failed: %m");
		iostat_close(mfd);
	}
}
