/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
/*
 * Uses the mountlist library in gnulib to map system device IDs and
 * system device names to mount entries.
 */
#include "cmogstored.h"

static pthread_mutex_t mnt_lock = PTHREAD_MUTEX_INITIALIZER;
/* global for the process */
static Hash_table *by_dev; /* maps (system) device IDs to a mount_entry */

struct mog_mntent {
	struct mount_entry *me;
	char util[8];
};

static void me_free(void *entry)
{
	struct mog_mntent *mntent = entry;
	struct mount_entry *me = mntent->me;

	free(me->me_devname);
	free(me->me_mountdir);
	if (me->me_type_malloced)
		free(me->me_type);
	free(me);
	free(mntent);
}

static size_t me_dev_hash(const void *entry, size_t tablesize)
{
	const struct mog_mntent *mntent = entry;

	return mntent->me->me_dev % tablesize;
}

static bool me_dev_cmp(const void *a, const void *b)
{
	const struct mog_mntent *mntent_a = a;
	const struct mog_mntent *mntent_b = b;

	return mntent_a->me->me_dev == mntent_b->me->me_dev;
}

static void mnt_atexit(void)
{
	hash_free(by_dev);
}

static void mnt_once(void)
{
	by_dev = hash_initialize(7, NULL, me_dev_hash, me_dev_cmp, me_free);
	if (!by_dev)
		mog_oom();
	atexit(mnt_atexit);
}

void mog_mnt_refresh(void)
{
	struct mount_entry *head = read_file_system_list(false);
	struct mog_mntent *mntent;
	struct stat sb;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );

	if (by_dev)
		hash_clear(by_dev);
	else
		mnt_once();

	while (head) {
		struct mount_entry *next = head->me_next;

		mntent = xmalloc(sizeof(struct mog_mntent));
		mntent->me = head;
		mntent->util[0] = '-';
		mntent->util[1] = 0;

		/* the device number may not have been populated, do it */
		if ((head->me_dev == (dev_t)-1) &&
		    (stat(head->me_mountdir, &sb) == 0))
			head->me_dev = sb.st_dev;

		switch (hash_insert_if_absent(by_dev, mntent, NULL)) {
		case 0: me_free(mntent);
		case 1: break;
		default: mog_oom();
		}
		head = next;
	}

	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
}

/*
 * Looks up a mount_entry by st_dev, returns NULL if nothing was found
 * Users may only acquire one mount entry at a time and MUST release it
 */
const struct mount_entry * mog_mnt_acquire(dev_t st_dev)
{
	struct mount_entry me = { .me_dev = st_dev };
	struct mog_mntent mntent = { .me = &me };
	struct mog_mntent *rv;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	rv = hash_lookup(by_dev, &mntent);

	/* user must release this via mog_mnt_release if non-NULL */
	if (rv) return rv->me;

	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
	return NULL;
}

/* releases the mount entry, allowing mog_mnt_acquire to be called again */
void mog_mnt_release(const struct mount_entry *me)
{
	struct mog_mntent mntent;
	const struct mog_mntent *check_me;
	union { const void *in; void *out; } deconst = { .in = me };

	mntent.me = deconst.out;
	check_me = hash_lookup(by_dev, &mntent);

	assert(check_me->me == me && "did not release acquired mount_entry");
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
}

struct mnt_update {
	const char *prefix;
	size_t prefixlen;
	const char *util;
};

static bool update_util_each(void *ent, void *upd)
{
	struct mog_mntent *mntent = ent;
	struct mnt_update *update = upd;
	const char *devname = mntent->me->me_devname;
	size_t devnamelen = strlen(devname);

	if (devnamelen >= update->prefixlen &&
	    memcmp(update->prefix, devname, update->prefixlen) == 0)
		memcpy(mntent->util, update->util, 8);
	return true; /* continue */
}

/* this is O(mountpoints) for now */
void mog_mnt_update_util(const char *devsuffix, const char *util, size_t len)
{
	struct mnt_update update;

	assert(len >= 8 && "len too small");
	update.prefix = xasprintf("/dev/%s", devsuffix);
	update.prefixlen = strlen(update.prefix);
	update.util = util;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	(void)hash_do_for_each(by_dev, update_util_each, &update);
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
	mog_free(update.prefix);
}

char * mog_mnt_fetch_util(dev_t st_dev, char *dst, size_t len)
{
	struct mount_entry me = { .me_dev = st_dev };
	struct mog_mntent mntent = { .me = &me };
	struct mog_mntent *tmp;

	assert(len >= 8 && "len too small for dst");

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	tmp = hash_lookup(by_dev, &mntent);
	if (tmp)
		memcpy(dst, tmp->util, 8);
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );

	return tmp ? dst : NULL;
}
