#include "cmogstored.h"

/* poll() is broken on some *BSDs for pipe */
#include <sys/select.h>

static int self_pipe[2];
static int notes[MOG_NOTIFY_MAX];
static time_t usage_file_updated_at;
static time_t usage_file_interval = 10;

void mog_notify_init(void)
{
	const char *interval = getenv("MOG_DISK_USAGE_FILE_INTERVAL");

	if (pipe2(self_pipe, O_NONBLOCK | O_CLOEXEC) < 0)
		die("failed to init self-pipe: %s\n", strerror(errno));

	if (interval) {
		int i = atoi(interval);

		if (i > 0)
			usage_file_interval = (time_t)i;
	}
}

static bool svc_mkusage_each(void *svc, void *cb)
{
	mog_svc_scandev((struct mog_svc *)svc, cb);

	return true;
}

static void global_mkusage(void)
{
	mog_svc_each(svc_mkusage_each, mog_dev_mkusage);
	usage_file_updated_at = time(NULL);
}

static void device_refresh(void)
{
	mog_svc_each(svc_mkusage_each, NULL);
}

static void drain_pipe(void)
{
	char buf[128];
	ssize_t r;

	do
		r = read(self_pipe[0], buf, sizeof(buf));
	while (r > 0 || (r < 0 && errno == EINTR));
	assert(r < 0 && errno == EAGAIN && "self-pipe read failed");
}

/* this is the main loop of cmogstored */
void mog_notify_wait(void)
{
	int rc;
	struct timeval tv;
	fd_set rfds;
	time_t next = usage_file_updated_at + usage_file_interval;
	time_t now = time(NULL);
	time_t timeout = next - now;

	rc = mog_sync_lock_test_and_set(&notes[MOG_NOTIFY_DEVICE_REFRESH], 0);
	if (next <= now) global_mkusage();
	if (rc) device_refresh();
	if (timeout <= 0) return;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	FD_ZERO(&rfds);
	FD_SET(self_pipe[0], &rfds);

	mog_intr_enable();
	rc = select(self_pipe[0] + 1, &rfds, NULL, NULL, &tv);
	mog_intr_disable();
	if (rc > 0)
		drain_pipe();
}

/* this is async-signal safe */
void mog_notify(enum mog_notification note)
{
	ssize_t w;

	switch (note) {
	case MOG_NOTIFY_DEVICE_REFRESH:
		(void)mog_sync_add_and_fetch(&notes[note], 1);
		break;
	/* support changing thread counts later... */
	case MOG_NOTIFY_SIGNAL: break;
	default: assert(0 && "bad note passed");
	}

retry:
	w = write(self_pipe[1], "^", 1);
	if (w >= 0) return;

	switch (errno) {
	case_EAGAIN: return; /* somebody already woke this up */
	case EINTR: goto retry; /* just in case... */
	}

	/*
	 * we're screwed anyways, at least try this even though syslog()
	 * isn't safe inside a signal handler.
	 */
	syslog(LOG_CRIT, "mog_notify write() to pipe failed: %m");
	abort();
}
