/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

static size_t dev_hash(const void *x, size_t tablesize)
{
	const struct mog_dev *dev = x;

	return dev->st_dev % tablesize;
}

static bool dev_cmp(const void *a, const void *b)
{
	const struct mog_dev *dev_a = a;
	const struct mog_dev *dev_b = b;

	return dev_a->st_dev == dev_b->st_dev;
}

static void svc_init_dev_hash(struct mog_svc *svc)
{
	if (svc->by_st_dev) {
		hash_clear(svc->by_st_dev);
		return;
	}
	svc->by_st_dev = hash_initialize(7, NULL, dev_hash, dev_cmp, free);
	if (!svc->by_st_dev) mog_oom();
}

int mog_svc_scandev(struct mog_svc *svc, mog_scandev_cb cb)
{
	struct dirent *ent;
	int rc = 0;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));
	svc_init_dev_hash(svc);
	rewinddir(svc->dir);
	while ((ent = readdir(svc->dir))) {
		unsigned long mog_devid;
		char *end;
		size_t len = strlen(ent->d_name);
		struct mog_dev *dev;

		if (len <= 3) continue;
		if (memcmp("dev", ent->d_name, 3) != 0) continue;

		mog_devid = strtoul(ent->d_name + 3, &end, 10);
		if (*end != 0) continue;
		if (mog_devid > 0xffffff) continue; /* MEDIUMINT in DB */

		dev = mog_dev_new(svc, (uint32_t)mog_devid);
		if (!dev) continue;

		if (cb) rc |= cb(dev, svc); /* mog_dev_mkusage */
		switch (hash_insert_if_absent(svc->by_st_dev, dev, NULL)) {
		case 0: free(dev);
		case 1: break;
		default: mog_oom(); /* -1 */
		}
	}
	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));

	return rc;
}

static bool write_devstats(void *entry, void *file)
{
	struct mog_dev *dev = entry;
	FILE *fp = file;
	char util[8];
	char *found = mog_mnt_fetch_util(dev->st_dev, util, sizeof(util));

	if (found)
		fprintf(fp, "%u\t%s\n", dev->devid, found);
	else
		fprintf(fp, "%u\t-\n", dev->devid);
	return true;
}

/* updates per-svc device stats from the global mount list */
static void devstats_update(struct mog_svc *svc)
{
	struct iovec *tmp = &svc->devstats;
	FILE *fp;
	char *data;

	assert(svc->by_st_dev && "need to scan devices first");
	free(tmp->iov_base);

	fp = open_memstream(&data, &tmp->iov_len);
	if (!fp) mog_oom();

	hash_do_for_each(svc->by_st_dev, write_devstats, fp);
	fprintf(fp, ".\n");

	CHECK(int, 0, fclose(fp));
	tmp->iov_base = data;
}

void mog_svc_devstats_subscribe(struct mog_mgmt *mgmt)
{
	struct mog_svc *svc = mgmt->svc;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));
	LIST_INSERT_HEAD(&svc->devstats_subscribers, mgmt, subscribed);
	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));
}

/* called while iterating through all mog_svc objects */
bool mog_svc_devstats_broadcast(void *ent, void *ignored)
{
	struct mog_svc *svc = ent;
	struct mog_mgmt *mgmt, *tmp;
	struct iovec iov;
	struct mog_fd *mfd;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));

	devstats_update(svc);
	LIST_FOREACH_SAFE(mgmt, &svc->devstats_subscribers, subscribed, tmp) {
		assert(mgmt->wbuf == NULL && "wbuf not null");
		iov = svc->devstats;
		mog_mgmt_writev(mgmt, &iov, 1);

		if (mgmt->wbuf == NULL) continue; /* success */

		LIST_REMOVE(mgmt, subscribed);
		mfd = mog_fd_of(mgmt);
		if (mgmt->wbuf == MOG_WR_ERROR) {
			assert(mgmt->rbuf == NULL && "would leak rbuf");
			mog_close(mfd->fd);
		} else { /* blocked on write */
			mog_idleq_push(mgmt->svc->queue, mfd, MOG_QEV_WR);
		}
	}

	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));

	return true;
}
